//-----------------------------------------------------------------------------
//  (c) 2011 by Basler Vision Technologies
//  Division: Machine Vision
//  Sample for optimizing GigE Packet Size and Inter-Packet Delay as well as splitting up the bandwidth
//  Author:  M.Binev
//-----------------------------------------------------------------------------

// This application optimizes the Packet Size, the Inter-Packet Delay and the bandwidth 
// for all cameras found in a network

#include <OptimizeGigESettings.h>
#include <StoreConfigurationSet.h>
#include <exception>

// Namespace for using cout
using namespace std;

// Extract the serial number out of a self-defined string
String_t extractSerNum (String_t& st) 
{
	size_t pos= st.find(':');
	String_t str = st.substr( pos + 2 );	
	return str;
}

// This function can be used to wait for user input at the end of the sample program.
void pressEnterToExit()
{
    //comment the following two lines to disable wait on exit here
    cerr << endl << "Press enter to exit." << endl;
	Sleep(5000);
    while( cin.get() != '\n');
}

// The main
int main (int argc, char* argv[])
{
    // Initialize the pylon runtime system
    PylonInitialize();
	// The file we are going to write in
	std::ofstream file;

    // The destructor of this local class ensures that PylonTerminate() will be called when the application terminates. 
	// The destructor will be called even when the application terminates due to an unhandled exception.
    struct Terminator
    {
        ~Terminator() { PylonTerminate(); }
    } terminator;

    try
    {
        // Get the transport layer factory
        CTlFactory& TlFactory = CTlFactory::GetInstance();

        // Create the transport layer object needed to enumerate or
        // create a camera object of type Pylon::CBaslerGigECamera::DeviceClass ()
        ITransportLayer *pTl = TlFactory.CreateTl (Pylon::CBaslerGigECamera::DeviceClass ());

		// Open the output file "optimizeGigE.txt"
		file.open(outfile); 
		if (!file)
		{ 
           cout << "Error opening file: " << outfile << endl;
		   file << "Error opening file: " << outfile << endl;
           exit(1);
		}

        // Exit application if the specific transport layer is not available
        if ( ! pTl )
        {
            cerr << "Failed to create transport layer!" << endl;
			file << "Failed to create transport layer!" << endl;
            return 1;
        }

        // Get all attached cameras and exit application if no camera is found
        DeviceInfoList_t devices;
        if ( 0 == pTl->EnumerateDevices (devices))
        {
            cerr << "No camera present!" << endl;
			file << "No camera present!" << endl;
            return 1;			
        }		
		
		// Create a MAP to hold a string of the NICs' IP addresses and a string of the cameras' IP addresses
		typedef std::vector<Pylon::String_t> StringVector_t;
		typedef std::map<Pylon::String_t, StringVector_t > StringToStringVectorMap_t;
		// The map
		StringToStringVectorMap_t  map;

		// Some additional variables
		bool splitBandwidth= true;
		// Consider a maximum of 20 cameras
		bool configurationNumber [20]; 
		// Initialize the boolean array
		for (int i= 0; i < 20; i++)
			configurationNumber [i]= true;
		int camerasNumber [20];

		// Fill up the map
		for (unsigned int i= 0; i < devices.size (); i++)
		{
			Pylon::CBaslerGigECamera Camera= pTl->CreateDevice(devices[i]);
			Camera.Open();
			Pylon::CBaslerGigECamera::StreamGrabber_t StreamGrabber  = Camera.GetStreamGrabber (0);
			StreamGrabber.Open();
			String_t str= Camera.DeviceModelName() + ", SN: " + Camera.DeviceID();			
			// For an already existing adapter IP address, we will only add the new camera to the vector.
			// If the adapter IP address was new, we would also add the adapter's IP address.
			map[ StreamGrabber.DestinationAddr()].push_back( str);
			StreamGrabber.Close();
			Camera.Close();
		}

		// Iterate through the map to print out camera(s) connected to the given adapter(s)
		for (StringToStringVectorMap_t::const_iterator kvp = map.begin();kvp!= map.end(); ++ kvp)
		{
			cout << "\nConnected camera(s) to Adapter: "  << kvp ->first << endl;
			file << "\nConnected camera(s) to Adapter: "  << kvp ->first << endl;			
			for(StringVector_t::const_iterator it = kvp->second.begin(); it != kvp->second.end(); ++it)
			{
				cout << "Camera: "  << *it  << std::endl;
				file << "Camera: "  << *it  << std::endl;
			}
		}	
		// Print out some stuff
		cout << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" << endl;
		file << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" << endl;
		cout << "I. This part will configure the camera(s) for a \"Peer-to-Peer\" connection, i.e. a single camera connected to a single network adapter." << endl;
		file << "I. This part will configure the camera(s) for a \"Peer-to-Peer\" connection, i.e. a single camera connected to a single network adapter." << endl;
		cout << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
		file << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;

		for (unsigned int camIx = 0; camIx < devices.size (); camIx++)
		{
			// Create the camera object of the first available camera.
			// The camera object is used to set and get all available camera features.
			Pylon::CBaslerGigECamera Camera= pTl->CreateDevice(devices[camIx]);
		
			uint32_t result = OptimizeGigESettings (Camera, 9000, file);
			if (result == 0)
			{
				cout << "Settings for Camera " << Camera.GetDeviceInfo ().GetFriendlyName () << " were modified:" << endl;
				file << "Settings for Camera " << Camera.GetDeviceInfo ().GetFriendlyName () << " were modified:" << endl;
				cout << "- Packet Size was set to : " << Camera.GevSCPSPacketSize.GetValue () << endl;
				file << "- Packet Size was set to : " << Camera.GevSCPSPacketSize.GetValue () << endl;
				cout << "- Inter-Packet Delay was set to : " << Camera.GevSCPD.GetValue () << endl << endl;
				file << "- Inter-Packet Delay was set to : " << Camera.GevSCPD.GetValue () << endl << endl;
				
				char key[256];
				cout << "Do you wish to store settings in a Configuration Set (y/n)?" << endl;
				file << "Do you wish to store settings in a Configuration Set (y/n)?" << endl;
				cin >> key;
				file << key << endl;
				bool check= false;
				// Check if the input is correct
				do {
					if (key[0] == 'y' || key[0] == 'Y' || key[0] == 'n' || key[0] == 'N') check= true; 
					else {
						check= false;
						cout << "Please enter \'y\' or \'n\'!" << endl;
						file << "Please enter \'y\' or \'n\'!" << endl;
						cin >> key;
						file << key << endl;
					}
				} while (!check);
				
				if (key[0] == 'y' || key[0] == 'Y')
				{
					cout << "Enter number of Configuration Set (1, 2 or 3)" << endl;
					file << "Enter number of Configuration Set (1, 2 or 3)" << endl;
					int setNum;
					char sNum[256];
					cin >> sNum;
					file << sNum << endl;					
					check= false;
					// Check if the input is correct
					do {
							setNum= atoi(sNum);
							if (setNum == 1 || setNum == 2 || setNum == 3) check= true; 
							else {
								check= false;
								cout << "Please enter 1, 2 or 3!" << endl;
								file << "Please enter 1, 2 or 3!" << endl;
								cin >> sNum;
								file << sNum << endl;
							}
					} while (!check);

					cout << "Should Configuration Set No. " << setNum << " become your Startup Set (y/n)?" << endl;
					file << "Should Configuration Set No. " << setNum << " become your Startup Set (y/n)?" << endl;
					cin >> key;
					file << key << endl;					
					check= false;
					// Check if the input is correct
					do {
						if (key[0] == 'y' || key[0] == 'Y' || key[0] == 'n' || key[0] == 'N') check= true; 
						else {
							check= false;
							cout << "Please enter \'y\' or \'n\'!" << endl;
							file << "Please enter \'y\' or \'n\'!" << endl;
							cin >> key;
							file << key << endl;
						}
					} while (!check);
					
					switch (setNum)
					{
					case 1:
							StoreToConfigSet (Camera, 
											   Camera.GevSCPSPacketSize.GetValue (), 
											   Camera.GevSCPD.GetValue (),
											   UserSetSelector_UserSet1, 
											   (key[0]=='y' || key[0] == 'Y'));
						break;
					case 2:
							StoreToConfigSet (Camera, 
											   Camera.GevSCPSPacketSize.GetValue (), 
											   Camera.GevSCPD.GetValue (),
											   UserSetSelector_UserSet2, 
											   (key[0]=='y' || key[0] == 'Y'));
						break;
					case 3:
							StoreToConfigSet (Camera, 
											   Camera.GevSCPSPacketSize.GetValue (), 
											   Camera.GevSCPD.GetValue (),
											   UserSetSelector_UserSet3, 
											   (key[0]=='y' || key[0] == 'Y'));
						break;
					default:
						{
							cerr << "Invalid Configuration Set!" << endl;
							file << "Invalid Configuration Set!" << endl;
						}
					}
				}
			}
			else if (result == 1)
			{
				cerr << "\nGrab timed out while calculating the Inter-Packet Delay!" << endl;
				file << "\nGrab timed out while calculating the Inter-Packet Delay!" << endl;
			}
			else
			{
				cerr << "\nError in grab while calculating the Inter-Packet Delay: " << result << endl;
				file << "\nError in grab while calculating the Inter-Packet Delay: " << result << endl;
			}			
		}
		// Print out some more stuff
		cout << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" << endl;
		file << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n" << endl;
		cout << "II. This part will configure any multiple cameras connected to the same network adapter(s) if needed." << endl;
		file << "II. This part will configure any multiple cameras connected to the same network adapter(s) if needed." << endl;
		cout << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
		file << "\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;

		int m= 0;
		// Iterate through the map to print out camera(s) connected to the given adapter(s)
		for (StringToStringVectorMap_t::const_iterator kvp = map.begin();kvp!= map.end(); ++ kvp, m++)
		{
			int n= 0;
			int64_t totalBandwidth= 0;
			for(StringVector_t::const_iterator it = kvp->second.begin(); it != kvp->second.end(); ++it, n++)
			{
				if (n >= 1) {
					if (n == 1) {
						cout << "\nThe following cameras are connected to the same adapter: "<< kvp ->first << "\nCamera SN: "  << extractSerNum((String_t)*kvp->second.begin()) << endl;
						file << "\nThe following cameras are connected to the same adapter: "<< kvp ->first << "\nCamera SN: "  << extractSerNum((String_t)*kvp->second.begin()) << endl;
						totalBandwidth= totalBandwidth + OptimizeBandwidth(extractSerNum((String_t)*kvp->second.begin()), 0, file);
						splitBandwidth= true;
					}
					cout << "Camera SN: "  << extractSerNum((String_t)(*it)) << endl;
					file << "Camera SN: "  << extractSerNum((String_t)(*it)) << endl;
					totalBandwidth= totalBandwidth + OptimizeBandwidth(extractSerNum((String_t)(*it)), 0, file);
					if ( (totalBandwidth > 112500000) && (n == kvp->second.size() - 1) ) {
						cout << "\nThe current Total Resulting Bandwidth for the above " << n+1 << " cameras is:\n" << totalBandwidth << " [Bytes/s]!\n";
						file << "\nThe current Total Resulting Bandwidth for the above " << n+1 << " cameras is:\n" << totalBandwidth << " [Bytes/s]!\n";
						cout << "\nYou MUST reduce it to about 112.5 [MB/s] by increasing the Inter-Packet Delay \nin order to avoid any image loss!\n\n"; 
						file << "\nYou MUST reduce it to about 112.5 [MB/s] by increasing the Inter-Packet Delay \nin order to avoid any image loss!\n\n";
						splitBandwidth= false;
					}
					else splitBandwidth= true;
				}
				configurationNumber[m]= splitBandwidth;
				camerasNumber[m]= n+1;
			}
		}

		// Do a pre-check whether we need to ask the user about splitting up the bandwidth
		bool var= true;
		for (int k= 0; k < 20; k++) {
				var= configurationNumber[k];
				if (!var) break;
		} 

		// Ask to split up the bandwidth
		if (!var) {
			cout << "\nDo you want this tool to split up the bandwidth between the cameras (y/n)?" << endl;
			file << "\nDo you want this tool to split up the bandwidth between the cameras (y/n)?" << endl;
			char key[256];
			cin >> key;
			file << key << endl;			
			bool check= false;
			// Check if the input is correct
			do {
				if (key[0] == 'y' || key[0] == 'Y' || key[0] == 'n' || key[0] == 'N') check= true; 
				else {
					check= false;
					cout << "Please enter \'y\' or \'n\'!" << endl;
					file << "Please enter \'y\' or \'n\'!" << endl;
					cin >> key;
					file << key << endl;
				}
			} while (!check);

			if (key[0] == 'y' || key[0] == 'Y') {
				int m= 0;
				// Iterate through the map
				for (StringToStringVectorMap_t::const_iterator kvp = map.begin();kvp!= map.end(); ++ kvp, m++)
				{
					int n= 0;
					int64_t totalBandwidth= 0;
					for(StringVector_t::const_iterator it = kvp->second.begin(); it != kvp->second.end(); ++it, n++)
					{
						if (n >= 1) {
							if (n == 1) {
								cout << "\nSplitting up the bandwidth for Camera SN: "  << extractSerNum((String_t)*kvp->second.begin()) << endl;
								file << "\nSplitting up the bandwidth for Camera SN: "  << extractSerNum((String_t)*kvp->second.begin()) << endl;
								OptimizeBandwidth(extractSerNum((String_t)*kvp->second.begin()), 112500000/camerasNumber[m], file);
							}
							cout << "\nSplitting up the bandwidth for Camera SN: "  << extractSerNum((String_t)(*it)) << endl;	
							file << "\nSplitting up the bandwidth for Camera SN: "  << extractSerNum((String_t)(*it)) << endl;
							OptimizeBandwidth(extractSerNum((String_t)(*it)), 112500000/camerasNumber[m], file);	
						}
					}
				}
			}
			else ;
		}
		else {
			cout << "\nNothing to do!" << endl;
			file << "\nNothing to do!" << endl;
		}			
	}
    catch (GenICam::GenericException &e)
    {
        // Error handling
        cerr << "An exception occurred!" << endl << e.GetDescription() << endl;
		file << "An exception occurred!" << endl << e.GetDescription() << endl;
		// Ask to exit
		pressEnterToExit();
        return 1;
    }

	cout << "\n\nThe complete configuration was stored in \"optimizeGigE.txt\"!" << endl;
	file << "\n\nThe complete configuration was stored in \"optimizeGigE.txt\"!" << endl;
	// Close the output file
	file.close();
	
	// Ask to exit
	pressEnterToExit();
    
	// Quit application
    return 0;
}
