//-----------------------------------------------------------------------------
//  (c) 2011 by Basler Vision Technologies
//  Division: Machine Vision
//  Sample for optimizing GigE Packet Size and Inter-Packet Delay as well as splitting up the bandwidth
//  Author:  M.Binev
//  Implementation: OptimizeGigESettings.cpp
//-----------------------------------------------------------------------------

#include <OptimizeGigESettings.h>

enum EControlState
{
	ToHigh,
	Correct,
	ToLow
};

class CGrabBuffer
{
    public:
        CGrabBuffer(const size_t ImageSize);
        ~CGrabBuffer();
        uint8_t* GetBufferPointer(void) { return m_pBuffer; }
        StreamBufferHandle GetBufferHandle(void) { return m_hBuffer; }
        void SetBufferHandle(StreamBufferHandle hBuffer) { m_hBuffer = hBuffer; };

    protected:
        uint8_t *m_pBuffer;
        StreamBufferHandle m_hBuffer;
};

// Constructor allocates the image buffer
CGrabBuffer::CGrabBuffer(const size_t ImageSize):
        m_pBuffer(NULL)
{
    m_pBuffer = new uint8_t[ ImageSize ];
    if (NULL == m_pBuffer)
    {
        GenICam::GenericException e("Not enough memory to allocate image buffer", __FILE__, __LINE__);
        throw e;
    }
}

// Freeing the memory
CGrabBuffer::~CGrabBuffer()
{
    if (NULL != m_pBuffer)
        delete[] m_pBuffer;
}

int64_t OptimizeBandwidth (Pylon::String_t& SN, int bandwidth, std::ostream& file) 
{
	// The current bandwidth
	int64_t currentBandwidth;	
	// Create a camera device by Serial Number
	IPylonDevice* device = CTlFactory::GetInstance().CreateDevice( CDeviceInfo().SetDeviceClass( BaslerGigEDeviceClass).SetSerialNumber( SN ));
	device->Open();
	// Get the camera control object
	INodeMap &Control = *device->GetNodeMap();
	const CIntegerPtr DeviceCurrentThroughput = Control.GetNode("GevSCDCT");
	// Get the current bandwidth needed
	currentBandwidth= DeviceCurrentThroughput->GetValue();
	const CIntegerPtr IPD = Control.GetNode("GevSCPD");
	int64_t inc= IPD->GetValue();
	int64_t max= IPD->GetMax();
	const CFloatPtr resFR= Control.GetNode("ResultingFrameRateAbs");
	
	if (bandwidth != 0) {
		
		do {
			inc+=32;
			if (inc <= max)
				IPD->SetValue(inc);
			else { 
				cout << "\n\nThe maximum Inter-Packet Delay for this camera was reached before reaching \nthe target bandwidth ( " << bandwidth << " [Bytes/s] )!\n\nEither consider updating the camera firmware or decreasing the bandwidth \nmanually, e.g. by enabling \"Enable Acquisition Frame Rate\" and setting \n\"Acquistion Frame Rate (Abs)\" to a lower value!\n\n"; 
				file << "\n\nThe maximum Inter-Packet Delay for this camera was reached before reaching \nthe target bandwidth ( " << bandwidth << " [Bytes/s] )!\n\nEither consider updating the camera firmware or decreasing the bandwidth \nmanually, e.g. by enabling \"Enable Acquisition Frame Rate\" and setting \n\"Acquistion Frame Rate (Abs)\" to a lower value!\n\n";
			}
			currentBandwidth= DeviceCurrentThroughput->GetValue();

		} while (currentBandwidth > bandwidth && inc <= max);

		cout << "\n\tThe resulting Inter-Packet Delay is: " << IPD->GetValue() << endl;
		file << "\n\tThe resulting Inter-Packet Delay is: " << IPD->GetValue() << endl;
		cout << "\tThe Resulting Frame Rate (Abs) is: " << resFR->GetValue() << endl; 
		file << "\tThe Resulting Frame Rate (Abs) is: " << resFR->GetValue() << endl;
	}

	device->Close();
	return currentBandwidth;
}

uint32_t OptimizeGigESettings (Pylon::CBaslerGigECamera &Camera, int64_t iInitialPacketSize, std::ostream& file)
{
    try
    {
		// Grab 2000 test images to check for any image faults
		int imagesToGrab= 2000;	
		// Get the first stream grabber object of the selected camera
		Pylon::CBaslerGigECamera::StreamGrabber_t StreamGrabber  = Camera.GetStreamGrabber (0);
		// Open the camera
		Camera.Open ();
		// Open the stream grabber
		StreamGrabber.Open ();
		// Load the default set before starting with the configuration
		Camera.UserSetSelector.SetValue(UserSetSelector_Default);
		Camera.UserSetLoad.Execute();

		/////////////////////////////////////////////////////////////////
		//	First task: Setting Packet Size to optimal value
		/////////////////////////////////////////////////////////////////

		cout << "\nUsing camera: " << Camera.GetDeviceInfo().GetFriendlyName() << endl;
		file << "\nUsing camera: " << Camera.GetDeviceInfo().GetFriendlyName() << endl;
		cout << "Currently used GigE driver: " << StreamGrabber.Type.ToString() << endl << endl;
		file << "Currently used GigE driver: " << StreamGrabber.Type.ToString() << endl << endl;

		// Check if the Performance Driver is available
		if (StreamGrabber.TypeIsWindowsIntelPerformanceDriverAvailable() == 0) {
			cout << "!!!\nYou are not using a recommended Intel Pro 1000 network adapter currently or the Basler Pylon Performance driver was not installed yet!\nHence, you may need to upgrade your hardware!\n!!!\n" << endl; 
			file << "!!!\nYou are not using a recommended Intel Pro 1000 network adapter currently or the Basler Pylon Performance driver was not installed yet!\nHence, you may need to upgrade your hardware!\n!!!\n" << endl;
		}

		// Set image format to Raw if it's a colour camera, otherwise to Mono8
		if (GenApi::IsAvailable (Camera.PixelFormat.GetEntryByName ("BayerGB8")))
			Camera.PixelFormat.SetValue (PixelFormat_BayerGB8);
		else if	(GenApi::IsAvailable (Camera.PixelFormat.GetEntryByName ("BayerGR8")))
			Camera.PixelFormat.SetValue (PixelFormat_BayerGR8);
		else if	(GenApi::IsAvailable (Camera.PixelFormat.GetEntryByName ("BayerBG8")))
			Camera.PixelFormat.SetValue (PixelFormat_BayerBG8);
		else if	(GenApi::IsAvailable (Camera.PixelFormat.GetEntryByName ("BayerRG8")))
			Camera.PixelFormat.SetValue (PixelFormat_BayerRG8);
		else
			Camera.PixelFormat.SetValue (PixelFormat_Mono8);

		// Set the AOI to maximum
		Camera.OffsetX.SetValue (0);
		Camera.OffsetY.SetValue (0);
		Camera.Width.SetValue (Camera.Width.GetMax());
		Camera.Height.SetValue (Camera.Height.GetMax());

		// Set the camera to single frame mode
		Camera.TriggerSelector.SetValue (TriggerSelector_AcquisitionStart);
		Camera.TriggerMode.SetValue (TriggerMode_Off);
		Camera.AcquisitionMode.SetValue (AcquisitionMode_SingleFrame);
		// The trigger source must be set to 'Software' for we'll use it later.
        Camera.TriggerSource.SetValue( TriggerSource_Software);

		// set exposure time to min., so it will not affect the frame rate
		Camera.ExposureMode.SetValue (ExposureMode_Timed);
		Camera.ExposureTimeRaw.SetValue (Camera.ExposureTimeRaw.GetMin ());

		// We are going to use Chunks, i.e. the CRC Checksum, in order to check the integrity of images
		// Enable chunks in general
        if (GenApi::IsWritable(Camera.ChunkModeActive))
        {
            Camera.ChunkModeActive.SetValue(true);
        }
        else
        {
            cerr << "!!!\nThe camera doesn't support chunk features!\nThe Firmware of your camera might be too old and you would need to update it!\n!!!" << endl;
			file << "!!!\nThe camera doesn't support chunk features!\nThe Firmware of your camera might be too old and you would need to update it!\n!!!" << endl;
            return 1;
        }
        
        // Enable crc chunks
        Camera.ChunkSelector.SetValue(ChunkSelector_PayloadCRC16);
        Camera.ChunkEnable.SetValue(true);


		// Create a image buffer
		const size_t ImageSize = (size_t) (Camera.PayloadSize.GetValue ());
		uchar8_t *pBuffer = new uchar8_t[ImageSize];

		// We won't use image buffers greater than ImageSize
		StreamGrabber.MaxBufferSize.SetValue (ImageSize);

		// We won't queue more than one image buffer at a time
		StreamGrabber.MaxNumBuffer.SetValue (1);

		// Begin of packet size calculating loop.
		// Increase the Packet Size, beginning with iInitialPacketSize, until we see
		// image corruptions or do not receive any frame anymore 
		EControlState PacketState; 
		int64_t PacketSize = iInitialPacketSize;
		int64_t LastDelta = PacketSize;
		
		cout << "Testing for the optimal Packet Size:\n" << endl;
		file << "Testing for the optimal Packet Size:\n" << endl;
		
		do {
			PacketSize -= (PacketSize%4); // Packetsize must always be a multiple of 4!
			if (PacketSize < 220)
				PacketSize= 220;
			
			Camera.GevSCPSPacketSize.SetValue (PacketSize);

			// Allocate all resources for grabbing. Critical parameters like image
			// size now must not be changed until FinishGrab() is called.
			StreamGrabber.PrepareGrab ();

			// Buffers used for grabbing must be registered at the stream grabber.
			// The registration returns a handle to be used for queuing the buffer
			const StreamBufferHandle hBuffer = StreamGrabber.RegisterBuffer (pBuffer, ImageSize);

			// Put buffer into the grab queue for grabbing
			StreamGrabber.QueueBuffer (hBuffer, NULL);

			// Let the camera acquire one single image (Acquisiton mode equals SingleFrame!)
			Camera.AcquisitionStart.Execute ();
						
			// Wait for grabbed image with timeout of 3000 ms
			if (StreamGrabber.GetWaitObject ().Wait (3000))
			{
				// Get the grab result from the grabber's result queue
				GrabResult Result;
				StreamGrabber.RetrieveResult (Result);

				if (Result.Succeeded ())
				{

					// Check to see if a buffer containing chunk data has been received
					if (PayloadType_ChunkData != Result.GetPayloadType())
					{
						cerr << "Unexpected payload type received" << endl;
						file << "Unexpected payload type received" << endl;
						return 1;
					}

					// Create ChunkParser
					IChunkParser* pChunkParser = Camera.CreateChunkParser();
					if (NULL == pChunkParser)
					{
						cerr << "There is no chunk parser available for this device type!" << endl;
						file << "There is no chunk parser available for this device type!" << endl;
						return 1;
					}

					// Attach the image buffer with chunk data to the parser. The parser extracts
					// the included data from the chunk.
					pChunkParser->AttachBuffer(Result.Buffer(), Result.GetPayloadSize());

					// Since we have activated the CRC feature, we can check
					// the integrity of the buffer first.					
					if (pChunkParser->HasCRC() && pChunkParser->CheckCRC() == false)
					{
						cerr << "\n!!!\nImage was damaged!\nThe CRC-Checksum of the grabbed image was incorrect!\nYou may have problems with the network card driver or your cable!\nUpgrading of your NIC and/or swapping the Ethernet cable with a (new) CAT6 one may be required!\n!!!\n" << endl;
						file << "\n!!!\nImage was damaged!\nThe CRC-Checksum of the grabbed image was incorrect!\nYou may have problems with the network card driver or your cable!\nUpgrading of your NIC and/or swapping the Ethernet cable with a (new) CAT6 one may be required!\n!!!\n" << endl;
						PacketState = ToHigh;						
					}
					else
					{
						if (PacketSize == iInitialPacketSize)
							PacketState = Correct;
						else
							PacketState = ToLow;
					}

					pChunkParser->DetachBuffer();
					// Destroy the chunk parser
					Camera.DestroyChunkParser(pChunkParser);			
				}
				else
				{
					PacketState = ToHigh;
				}
			}
			else 
			{
				PacketState = ToHigh;
			}

			switch (PacketState)
			{
			case ToHigh:
					// Packet Size too high? Trying with half of former size
					LastDelta /= 2;
					PacketSize -= LastDelta;
				break;
			case ToLow:
					// Packet Size too low? Adding half of former size to packetsize
					LastDelta /= 2;
					PacketSize += LastDelta;
				break;
			}

			if (LastDelta == 0)
				PacketState = Correct;
			
			// Get the pending buffer back (It is not allowed to deregister
			// buffers when they are still queued)
			StreamGrabber.CancelGrab();

			// Get all buffers back
			for ( GrabResult r; StreamGrabber.RetrieveResult( r ););

			// It is necessary to deregister buffer before freeing the memory
			StreamGrabber.DeregisterBuffer( hBuffer );

			// Free all resources used for grabbing
			StreamGrabber.FinishGrab();

		} while (PacketState != Correct);

		// Free memory of image buffer
		delete[] pBuffer; 

		if (PacketSize <= 1504) 
		{
			PacketSize= 1500;
			Camera.GevSCPSPacketSize.SetValue (PacketSize);
			cout << "!!!\nYour network adapter does not support any Jumbo Frames or they were not enabled yet! If Jumbo Frames were not supported at all or their size was smaller \nthan 9 KByte, you may need to upgrade your PC with a recommended Intel Pro 1000 network adapter!\n!!!\n" << endl;
			file << "!!!\nYour network adapter does not support any Jumbo Frames or they were not enabled yet! If Jumbo Frames were not supported at all or their size was smaller \nthan 9 KByte, you may need to upgrade your PC with a recommended Intel Pro 1000 network adapter!\n!!!\n" << endl;
		}

		cout << "The optimal Packet Size is: " << PacketSize << endl;
		file << "The optimal Packet Size is: " << PacketSize << endl;

		/////////////////////////////////////////////////////////////////
		//	Second task: Setting Inter-Packet Delay to optimal value
		/////////////////////////////////////////////////////////////////

		// set AOI to max.
		Camera.Width.SetValue (Camera.Width.GetMax ());
		Camera.Height.SetValue (Camera.Height.GetMax ());

		// Disable chunks
		Camera.ChunkModeActive.SetValue(false);

		// Allocate all resources for grabbing. Critical parameters like image
		// size now must not be changed until FinishGrab() is called.
		StreamGrabber.PrepareGrab ();

		// Buffers used for grabbing must be registered at the stream grabber.
        // The registration returns a handle to be used for queuing the buffer.
        std::vector<CGrabBuffer*> BufferList;
		int c_nBuffers= 1;
        for (int i = 0; i < c_nBuffers; ++i)
        {
            CGrabBuffer *pGrabBuffer = new CGrabBuffer(ImageSize);
            pGrabBuffer->SetBufferHandle(StreamGrabber.RegisterBuffer(
                                             pGrabBuffer->GetBufferPointer(), ImageSize));

            // Put the grab buffer object into the buffer list
            BufferList.push_back(pGrabBuffer);
        }

        for (std::vector<CGrabBuffer*>::const_iterator x = BufferList.begin(); x != BufferList.end(); ++x)
        {
            // Put buffer into the grab queue for grabbing
            StreamGrabber.QueueBuffer((*x)->GetBufferHandle(), NULL);
        }
	
		// Begin of Inter-Packet Delay calculating loop.
		// Decrease Inter-Packet Delay, beginning with max. delay 
		// until the frame rate does not increase anymore 
		int64_t ipDelay = Camera.GevSCPD.GetMax ();
		double timeout;
		Camera.GevSCPD.SetValue (Camera.GevSCPD.GetMin());
		int orgFramerate= (int)Camera.ResultingFrameRateAbs.GetValue ();
		double orgFramerate2= Camera.ResultingFrameRateAbs.GetValue ();

		// This check will find out the rough IPD value that we will use in order to find the optimal IPD,
		// i.e. the one that would start to decrease the Resulting Frame Rate (Abs).
		do {
			Camera.GevSCPD.SetValue (ipDelay);
			timeout= Camera.ResultingFrameRateAbs.GetValue ();
			if ((int)timeout < orgFramerate)
				ipDelay=ipDelay/2;
			
		} while ( timeout < orgFramerate);

		// This check will find out the optimal IPD value that would start to decrease the Resulting Frame Rate (Abs).
		do {
			Camera.GevSCPD.SetValue (ipDelay);
			timeout= Camera.ResultingFrameRateAbs.GetValue ();			
			if ((int)timeout < orgFramerate2)
				ipDelay+=16;
			else
				ipDelay-=16;
			
		} while ( timeout >= orgFramerate2);

		Camera.GevSCPD.SetValue (ipDelay);
		cout << "\nThe optimal Inter-Packet Delay that will not decrease \nthe Resulting Frame Rate (Abs)= " << Camera.ResultingFrameRateAbs.GetValue() << " [fps] is: " << ipDelay << endl << endl;
		file << "\nThe optimal Inter-Packet Delay that will not decrease \nthe Resulting Frame Rate (Abs)= " << Camera.ResultingFrameRateAbs.GetValue() << " [fps] is: " << ipDelay << endl << endl;
				
		// Grab images in order to check if no Packets and/or Frames are going lost.
		cout << "Grabbing test images in order to check if any \nPackets or Frames are going lost:" << endl;
		file << "Grabbing test images in order to check if any \nPackets or Frames are going lost:" << endl;

		// Switch to Continuous acquisition mode with Software Trigger,
		// because it is more efficient while grabbing images
		Camera.TriggerMode.SetValue(TriggerMode_Off);
		Camera.AcquisitionMode.SetValue(AcquisitionMode_Continuous);

		int count= 0;
		int lostPackets= 0;
		int lastLP= 0;
		int lostFrames= 0;
		int lastLF= 0;
		int resendPackets= 0;
		int lastRP= 0;
		int64_t orgIpDelay= ipDelay;

		// Let the camera acquire continuously
		// (Acquisiton mode equals Continuous!)
		Camera.AcquisitionStart.Execute ();
		
		do {
			count++;
			if (count%5 == 0)
				cout << "\rTest image: " << count;

			Camera.GevSCPD.SetValue (ipDelay);			
			
			// Wait for grabbed image with timeout of 3 seconds
			if (StreamGrabber.GetWaitObject ().Wait(3000))
			{
				// Get the grab result from the grabber's result queue
				GrabResult Result;
				StreamGrabber.RetrieveResult( Result );
				lostPackets= (int)StreamGrabber.Statistic_Failed_Packet_Count();
				lostFrames= (int)StreamGrabber.Statistic_Failed_Buffer_Count();
				resendPackets= (int)StreamGrabber.Statistic_Resend_Packet_Count();

				// Do nothing if everything was ok
				if (lostPackets == lastLP && lostFrames == lastLF && resendPackets== lastRP);
				// Increase the IPD if a packet or a frame was lost or a packet resend occured.
				else {
					cout << "\nResend or image loss occured!\nThe Inter-Packet Delay will be increased!\n";
					ipDelay+=64;
				}
				
				lastLP= lostPackets;
				lastLF= lostFrames;
				lastRP= resendPackets;
		
				// Reuse the buffer
				StreamGrabber.QueueBuffer(Result.Handle(), NULL);
			}
			else 
			{
				// Timeout 
				return 1;
			}
		} while ( count < imagesToGrab );

		// Set the last used IPD
		Camera.GevSCPD.SetValue (ipDelay);
		if (orgIpDelay < ipDelay) 
		{
			orgIpDelay= (ipDelay - orgIpDelay)/64;
			file << "\nResend or image loss occured " << orgIpDelay << " times!\nThe Inter-Packet Delay was increased by totally " << orgIpDelay*64 << " [ticks]!" << endl;
		}
		cout << "\n\nThe final Inter-Packet Delay is: " << ipDelay << endl;
		file << "\n\nThe final Inter-Packet Delay is: " << ipDelay << endl;
		cout << "The Resulting Frame Rate (Abs) is: " << Camera.ResultingFrameRateAbs.GetValue() << endl << endl;
		file << "The Resulting Frame Rate (Abs) is: " << Camera.ResultingFrameRateAbs.GetValue() << endl << endl;

		// Stop the image acquisition
		Camera.AcquisitionStop.Execute();
		Camera.TriggerMode.SetValue(TriggerMode_Off);

		// Get the pending buffer back (It is not allowed to deregister
		// buffers when they are still queued)
		StreamGrabber.CancelGrab();

		// Get all buffers back
		for ( GrabResult r; StreamGrabber.RetrieveResult( r ););

		// You must deregister the buffers before freeing the memory
        for (std::vector<CGrabBuffer*>::iterator it = BufferList.begin(); it != BufferList.end(); it++)
        {
            StreamGrabber.DeregisterBuffer((*it)->GetBufferHandle());
            delete *it;
        }

		// Free all resources used for grabbing
		StreamGrabber.FinishGrab();

		// Close stream grabber
		StreamGrabber.Close();

		// Close camera
		Camera.Close();
	}
    catch (GenICam::GenericException &e)
    {
        throw e;
    }
	return 0;
}